# -*- coding: utf-8 -*-

# Copyright (c) 2018-2022 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

# The source code contains parts copied and modified from Spyder project:
# https://github.com/spyder-ide/spyder
# See below for the original copyright notice.

#
# Copyright (c) Spyder Project Contributors
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from types import ModuleType
import json
import ast

import spyder_kernels
from spyder_kernels.console.kernel import SpyderKernel

from .utility.tupleencoder import hinted_tuple_hook
from .utility.typeutil import (
    is_instance_of,
    is_numpy_number, numpy_to_py)

_spykern_ver = tuple(int(i) for i in spyder_kernels.__version__.split(".")[:3])

class ModelxKernel(SpyderKernel):

    def __init__(self, *args, **kwargs):
        super(ModelxKernel, self).__init__(*args, **kwargs)

        for call_id, handler in [
            ('mx_get_modellist', self.mx_get_modellist),
            ('mx_get_adjacent', self.mx_get_adjacent),
            ('mx_new_model', self.mx_new_model),
            ('mx_read_model', self.mx_read_model),
            ('mx_del_object', self.mx_del_object),
            ('mx_del_model', self.mx_del_model),
            ('mx_write_model', self.mx_write_model),
            ('mx_import_names', self.mx_import_names),
            ('mx_get_value', self.mx_get_value),
            ('mx_get_node', self.mx_get_node),
            ('mx_get_attrdict', self.mx_get_attrdict),
            ('mx_get_value_info', self.mx_get_value_info)

        ]:
            self.frontend_comm.register_call_handler(
                call_id,
                handler
            )

    def get_modelx(self):
        from modelx.core import mxsys
        return mxsys

    def mx_new_model(self, name=None, define_var=False, varname=''):
        import modelx as mx
        model = mx.new_model(name)
        if define_var:
            self._define_var(model, varname)
        self.send_mx_msg("mxupdated")

    def mx_read_model(self, modelpath, name, define_var, varname):
        import modelx as mx
        model = mx.read_model(modelpath, name)
        if define_var:
            self._define_var(model, varname)
        self.send_mx_msg("mxupdated")

    def mx_write_model(self, model, modelpath, backup, zipmodel):
        import modelx as mx
        if zipmodel:
            mx.zip_model(mx.get_models()[model], modelpath, backup)
        else:
            mx.write_model(mx.get_models()[model], modelpath, backup)

    def mx_new_space(self, model, parent, name, bases, define_var, varname):
        import modelx as mx

        model = self._get_or_create_model(model)

        if parent:
            parent = mx.get_object(parent)
        else:
            parent = model

        if not name:
            name = None

        if bases:
            bases = [model._get_from_name(b.strip()) for b in bases.split(",")]
        else:
            bases = None

        space = parent.new_space(name=name, bases=bases)
        if define_var:
            self._define_var(space, varname)

        self.send_mx_msg("mxupdated")

    def mx_del_object(self, parent, name):
        import modelx as mx
        mx.get_object(parent).__delattr__(name)
        self.send_mx_msg("mxupdated")

    def mx_del_model(self, name):
        import modelx as mx
        mx.get_models()[name].close()
        self.send_mx_msg("mxupdated")

    def _define_var(self, obj, varname=None, replace_existing=True):

        name = varname or obj.name

        if _spykern_ver > (2, 2):
            glbs = self.shell.user_ns
        else:
            glbs = self._mglobals()

        if replace_existing or name not in glbs:
            glbs[name] = obj
            return True
        else:
            return False

    def mx_import_names(self, fullname,
                        import_selected,
                        import_children, replace_existing):
        import modelx as mx
        from modelx.core.space import ItemSpace
        from modelx.core.spacecontainer import BaseSpaceContainer
        from modelx.core.space import BaseSpace
        from modelx.core.reference import ReferenceProxy

        obj = mx.get_object(fullname, as_proxy=True)

        if import_selected:  # Retrieve non item parent
            parent = obj
            while isinstance(parent, mx.core.space.ItemSpace):
                parent = obj.parent

            if isinstance(parent, ReferenceProxy):
                self._define_var(parent.value,
                                 varname=parent.name,
                                 replace_existing=replace_existing)
            else:
                self._define_var(parent, replace_existing=replace_existing)

        if import_children and isinstance(obj, BaseSpaceContainer):
            for child in obj.spaces.values():
                self._define_var(child, replace_existing=replace_existing)

            if isinstance(obj, BaseSpace):
                for child in obj.cells.values():
                    self._define_var(child, replace_existing=replace_existing)

            for name, child in obj.refs.items():
                self._define_var(child,
                                 varname=name,
                                 replace_existing=replace_existing)

        self.send_mx_msg("mxupdated")

    def mx_new_cells(self, model, parent, name, define_var, varname):
        """
        If name is blank and formula is blank, cells is auto-named.
        If name is blank and formula is func def, name is func name.
        If name is given and formula is blank, formula is lambda: None.
        If name is given and formula is given, formula is renamed to given name.

        Args:
            model: model name or blank
            parent: parent named id or blank
            name: cells name or blank
            formula: function def or lambda expression blank
        """
        model = self._get_or_create_model(model)
        if parent:
            parent = model._get_from_name(parent)
        else:
            parent = model.cur_space() if model.cur_space() else model.new_space()

        if not name:
            name = None

        ns = self.shell.user_ns if _spykern_ver > (2, 2) else self._mglobals()
        if "__mx_temp" in ns:
            formula = ns["__mx_temp"]
            del ns["__mx_temp"]
        else:
            formula = None

        cells = parent.new_cells(
            name=name,
            formula=formula
        )
        if define_var:
            self._define_var(cells, varname)
        self.send_mx_msg("mxupdated")

    def mx_set_formula(self, fullname):
        import modelx as mx

        obj = mx.get_object(fullname)
        ns = self.shell.user_ns if _spykern_ver > (2, 2) else self._mglobals()
        if "__mx_temp" in ns:
            formula = ns["__mx_temp"]
            del ns["__mx_temp"]
        else:
            formula = None

        obj.set_formula(formula)
        self.send_mx_msg("mxupdated")

    def _get_or_create_model(self, model):
        """
        if model is blank, current model is used if exits, otherwise new model
        is created.
        """
        import modelx as mx

        if model:
            return mx.get_models()[model]
        else:
            return mx.cur_model() if mx.cur_model() else mx.new_model()

    def mx_get_object(self, msgtype, fullname=None, attrs=None, recursive=False):
        # TODO: Replace with mx_get_attrdict

        import modelx as mx
        if fullname is None:
            obj = mx.cur_model()
        else:
            try:
                obj = mx.get_object(fullname, as_proxy=True)
            except NameError:
                obj = None

        if obj is not None:
            data = obj._get_attrdict(attrs, recursive=recursive)
        else:
            data = None

        self.send_mx_msg(msgtype, data=data)


    def mx_get_attrdict(self, fullname=None, attrs=None, recursive=False):

        import modelx as mx
        if fullname is None:
            obj = mx.cur_model()
        else:
            try:
                obj = mx.get_object(fullname, as_proxy=True)
            except NameError:
                obj = None

        if obj is not None:
            data = obj._get_attrdict(attrs, recursive=recursive)
        else:
            data = None

        return data

    def mx_get_modellist(self):
        """Returns a list of model info.

         Returns a list of dicts of basic model attributes.
         The first element of the list is the current model info.
         None if not current model is set.
         """

        import modelx as mx
        from modelx.core.base import Interface

        data = [m._get_attrdict(recursive=False) for m in mx.get_models().values()]

        if mx.cur_model():
            cur = mx.cur_model()._get_attrdict(recursive=False)
        else:
            cur = None

        data.insert(0, cur)

        return data

    def mx_get_codelist(self, fullname):
        import modelx as mx

        try:
            obj = mx.get_object(fullname)
            data = obj._get_attrdict(['formula'])
        except:
            data = None

        self.send_mx_msg('codelist', data=data)

    def mx_get_evalresult(self, msgtype, data):

        begstr = "analyze_"
        endstr = "_setnode"
        if msgtype[:len(begstr)] == begstr and msgtype[-len(endstr):] == endstr:
            if "value" in data:
                data["value"] = self._to_sendval(data["value"])

        # The code below is based on SpyderKernel.get_value
        try:
            self.send_mx_msg(msgtype, data=data)
        except:
            # * There is no need to inform users about
            #   these errors.
            # * value = None makes Spyder to ignore
            #   petitions to display a value
            self.send_mx_msg(msgtype, data=None)

        self._do_publish_pdb_state = False

    def mx_get_node(self, msgtype, fullname: str, argstr: str):
        import modelx as mx
        from modelx.core.reference import ReferenceProxy
        from modelx.core.base import Interface

        args = ast.literal_eval(argstr)
        obj = mx.get_object(fullname, as_proxy=True)

        node = obj.node(*args)
        data = node._get_attrdict(recursive=False, extattrs=['formula'])
        data["value"] = self._to_sendval(data["value"])

        return data

    def mx_get_adjacent(self, msgtype, obj: str,
                        jsonargs: str, adjacency: str):

        import modelx as mx
        from modelx.core.base import Interface

        args = json.loads(jsonargs, object_hook=hinted_tuple_hook)
        node = mx.get_object(obj).node(*args)
        nodes = getattr(node, adjacency)
        attrs = [node._get_attrdict(
            recursive=False, extattrs=['formula']) for node in nodes]

        for node in attrs:
            node["value"] = self._to_sendval(node["value"])

        return attrs

    def mx_get_value_info(self, model: str):

        import modelx as mx

        values = mx.get_models()[model]._get_assoc_values()

        i = 0
        size = len(values)

        while i < size:
            val = values[i]

            val["value"] = self._to_sendval(val["value"])
            if val["spec"]:
                val["spec"] = val["spec"]._get_attrdict()
                if "value" in val["spec"]:
                    val["spec"]["value"] = val["value"]

            val["refs"] = list(ref._get_attrdict() for ref in val["refs"])
            i += 1

        return values

    def _to_sendval(self, value):
        import modelx
        mxver = tuple(int(i) for i in modelx.__version__.split(".")[:3])
        from modelx.core.cells import Interface

        if mxver < (0, 18, 0):
            from modelx.io.baseio import BaseDataClient as dataspec
        else:
            from modelx.io.baseio import BaseDataSpec as dataspec

        if isinstance(value, (Interface, str, dataspec,
                              ModuleType)):
            return repr(value)

        elif any(
                is_instance_of(value, c, "pandas") for c in [
                    "DataFrame", "Index", "Series"
                ]):
            return "Type: " + value.__class__.__name__

        elif any(
            is_instance_of(value, c, "numpy") for c in [
                "ndarray", "MaskedArray"
            ]):
            return "Type: " + value.__class__.__name__

        elif isinstance(value, (list, set, tuple, dict)):
            return "Type: " + value.__class__.__name__

        elif is_numpy_number(value):
            return numpy_to_py[type(value).__name__](value)
        else:
            return value

    def mx_get_value(self, msgtype, fullname: str, argstr: str, calc: bool):
        """Get value of modelx object

        Returns a pair of the value and bool to indicate if the value is just
        calculated

        """
        import modelx as mx
        from modelx.core.reference import ReferenceProxy
        from modelx.core.base import Interface

        args = ast.literal_eval(argstr)
        obj = mx.get_object(fullname, as_proxy=True)
        if isinstance(obj, ReferenceProxy):
            value = [mx.get_object(fullname), False]
        elif isinstance(obj, Interface):
            if args in obj:
                value = [obj(*args), False]
            else:
                if calc:
                    value = [obj(*args), True]
                else:
                    raise KeyError("value for %s not found" % argstr)

        return value

    def send_mx_msg(self, mx_msgtype, content=None, data=None):
        """
        Publish custom messages to the Spyder frontend.

        This code is modified from send_spyder_msg in spyder-kernel v0.2.4

        Parameters
        ----------

        mx_msgtype: str
            The spyder message type
        content: dict
            The (JSONable) content of the message
        data: any
            Any object that is serializable by cloudpickle (should be most
            things). Will arrive as cloudpickled bytes in `.buffers[0]`.
        """
        import cloudpickle

        if content is None:
            content = {}
        content['mx_msgtype'] = mx_msgtype
        self.session.send(
            self.iopub_socket,
            'modelx_msg',
            content=content,
            buffers=[cloudpickle.dumps(data, protocol=2)],
            parent=self._parent_header)




# -*- coding: utf-8 -*-

# Copyright (c) 2018-2019 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

import json

# Code below modified from
# https://stackoverflow.com/questions/15721363/preserve-python-tuples-with-json


class TupleEncoder(json.JSONEncoder):
    """JSON encoder preserving Python tuples"""
    def encode(self, obj):
        return super(TupleEncoder, self).encode(hint_tuples(obj))


def hint_tuples(item):
    if isinstance(item, tuple):
        return {'__tuple__': True, 'items': hint_tuples(list(item))}
    if isinstance(item, list):
        return [hint_tuples(e) for e in item]
    if isinstance(item, dict):
        return {key: hint_tuples(value) for key, value in item.items()}
    else:
        return item


def hinted_tuple_hook(obj):
    if '__tuple__' in obj:
        return tuple(obj['items'])
    else:
        return obj


def test_tuple_encoder():

    sample = (1, 2, '藍上夫', (3, 4.33), [5, 6, (7, 8, [9, 10], 'ABC')])

    enc = TupleEncoder(ensure_ascii=False)
    encoded = enc.encode(sample)
    print(encoded)

    decoded = json.loads(encoded, object_hook=hinted_tuple_hook)
    print("%s == %s" % (sample, decoded))
    assert sample == decoded


if __name__ == "__main__":
    test_tuple_encoder()



# -*- coding: utf-8 -*-

# Copyright (c) 2018-2022 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

# The source code contains parts copied and modified from Spyder project:
# https://github.com/spyder-ide/spyder
# See below for the original copyright notice.

#
# Copyright (c) Spyder Project Contributors
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import sys
import os.path


def main():
    """Copied and modified from spyder.utils.ipython.start_kernel.py(v3.2.8)"""

    # Remove this module's path from sys.path:
    try:
        sys.path.remove(os.path.dirname(__file__))
    except ValueError:
        pass

    try:
        locals().pop('__file__')
    except KeyError:
        pass
    __doc__ = ''
    __name__ = '__main__'

    # Add current directory to sys.path (like for any standard Python interpreter
    # executed in interactive mode):
    sys.path.insert(0, '')

    from spyder_kernels.console.start import import_spydercustomize
    # Import our customizations into the kernel
    import_spydercustomize()

    # Fire up the kernel instance.
    from ipykernel.kernelapp import IPKernelApp

    from spyder_kernels.console.start import kernel_config, varexp
    from spymx_kernels.mxkernel import ModelxKernel

    kernel = IPKernelApp.instance()
    kernel.kernel_class = ModelxKernel
    try:
        kernel.config = kernel_config()
    except:
        pass
    kernel.initialize()

    # The comment below seems to be false for at least Spyder v3.3.2
    # __IPYTHON___ is indeed defined when modelx system is created.
    #
    # (Setup modelx for IPython here because
    # __IPYTHON__ is not defined before kernel.initialize())
    #
    # kernel.kernel.get_modelx().setup_ipython()

    # Set our own magics
    kernel.shell.register_magic_function(varexp)

    # Set Pdb class to be used by %debug and %pdb.
    # This makes IPython consoles to use the class defined in our
    # sitecustomize instead of their default one.
    import pdb
    kernel.shell.InteractiveTB.debugger_cls = pdb.Pdb

    # Start the (infinite) kernel event loop.
    kernel.start()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Kernels Contributors
#
# Licensed under the terms of the MIT License
# (see spyder_kernels/__init__.py for details)
# -----------------------------------------------------------------------------

"""Jupyter Kernels for the Spyder consoles."""

# Standard library imports
import ast
import io
import os

# Third party imports
from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))

with io.open('README.md', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()


def get_version(module='spymx_kernels'):
    """Get version."""
    with open(os.path.join(HERE, module, '_version.py'), 'r') as f:
        data = f.read()
    lines = data.split('\n')
    for line in lines:
        if line.startswith('VERSION_INFO'):
            version_tuple = ast.literal_eval(line.split('=')[-1].strip())
            version = '.'.join(map(str, version_tuple))
            break
    return version


REQUIREMENTS = [
    'spyder-kernels>=1.8.1'
]

TEST_REQUIREMENTS = [
    # 'codecov',
    # 'cython',
    # 'dask[distributed]',
    # 'flaky',
    # 'matplotlib',
    # 'mock',
    # 'numpy',
    # 'pandas',
    # 'pytest',
    # 'pytest-cov',
    # 'scipy',
    # 'xarray',
    # 'pillow',
]

setup(
    name='spymx-kernels',
    version=get_version(),
    keywords='modelx spyder jupyter kernel ipython console',
    url='https://github.com/fumitoh/spymx-kernels',
    download_url="https://pypi.org/project/spymx-kernels",
    license='LGPLv3',
    author='Fumito Hamamura',
    author_email="fumito.ham@gmail.com",
    description="Jupyter kernels for Spyder plugin for modelx",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['docs', '*tests']),
    install_requires=REQUIREMENTS,
    extras_require={'test': TEST_REQUIREMENTS},
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Jupyter',
        'Framework :: IPython',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Interpreters',
    ]
)






def is_instance_of(obj, class_: str, module: str):
    return is_class_of(type(obj), class_, module)


def is_class_of(type_, class_: str, module: str):
    """Check type without importing the type"""
    if type_.__name__ == class_:
        type_module = type_.__module__.split(".")[0]
        if type_module == module:
            return True

    return False


_numpy_number_types = [
    "bool_",
    "int_",
    "intc",
    "intp",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "float_",
    "float16",
    "float32",
    "float64",
    "complex_",
    "complex64",
    "complex128"
]

numpy_to_py = {
    "bool_": bool,
    "int_": int,
    "intc": int,
    "intp": int,
    "int8": int,
    "int16": int,
    "int32": int,
    "int64": int,
    "uint8": int,
    "uint16": int,
    "uint32": int,
    "uint64": int,
    "float_": float,
    "float16": float,
    "float32": float,
    "float64": float,
    "complex_": complex,
    "complex64": complex,
    "complex128": complex
}

def is_numpy_number(obj):
    """Check if numpy number

        bool_,
        int_,
        intc,
        intp,
        int8,
        int16,
        int32,
        int64,
        uint8,
        uint16,
        uint32,
        uint64,
        float_,     alias for float64
        float16,
        float32,
        float64,    subclass of Python float
        complex_,   alias for complex128
        complex64,
        complex128  subclass of Python complex

    """
    type_ = type(obj)
    type_name = type_.__name__
    type_module = type_.__module__.split(".")[0]

    if type_module == "numpy":
        if type_name in _numpy_number_types:
            return True

    return False



